(defun myash (m)
  (if (> m 65536)
      (throw 'fail (list 'toobigtoshift m))
    (ash 1 m)))
					; assume HF sets represented as numbers (Ackermann)
(defun num-to-list (n &optional (i 0))
  (if (> n 0)
      (if (= (logand n 1) 1)
	  (cons i (num-to-list (ash n -1) (1+ i)))
	(num-to-list (ash n -1) (1+ i)))
    nil))

(defun list-to-num (l)
  (if l
      (logior (myash (car l)) (list-to-num (cdr l)))
    0))

(defun mem (n m)
  (member n (num-to-list m) :test #'equal))

(defun subq (n m)
  (subsetp (num-to-list n) (num-to-list m) :test #'equal))

(defun disj (i j) (not (intersection (num-to-list i) (num-to-list j) :test #'equal)))

(defun powerlist (l)
  (if l
      (let ((r (powerlist (cdr l))))
	(append r (mapcar #'(lambda (x) (cons (car l) x)) r)))
    (list nil)))

(defun power (n)
  (if (> n 1023)
      (throw 'fail 'fail)
    (list-to-num (mapcar #'list-to-num (powerlist (num-to-list n))))))

(defun union1 (n)
  (let ((rl nil))
    (dolist (a (num-to-list n))
      (setq rl (append (num-to-list a) rl)))
    (list-to-num rl)))

(defun sing (n)
  (myash n))

(defun upair (n m)
  (logior (myash n) (myash m)))

(defun binunion (n m)
  (list-to-num (append (num-to-list n) (num-to-list m))))

(defun binintersect (n m)
  (list-to-num (intersection (num-to-list n) (num-to-list m) :test #'equal)))

(defun setminus (n m)
  (list-to-num (set-difference (num-to-list n) (num-to-list m) :test #'equal)))

(defun setadjoin (n m)
  (logior n (myash m)))

(defun bd-forall (n p)
  (let ((l (num-to-list n)))
    (dolist (a l t)
      (unless (funcall p a)
	(return nil)))))
  
(defun bd-exists (n p)
  (let ((l (num-to-list n)))
    (dolist (a l nil)
      (unless (funcall p a)
	(return t)))))
  
(defun power-closed (n)
  (let ((l (num-to-list n)))
    (dolist (a l t)
      (unless (member (power a) l :test #'equal)
	(return nil)))))

(defun TransSet (n)
  (bd-forall n #'(lambda (a) (subq a n))))
	     
(defun ordinal (n)
  (and (transset n)
       (bd-forall n #'ordinal)))

					; since everything is finite here, nat-p and ordinal are the same
(defun nat-p (n)
  (ordinal n))

					; this version actually computes the number or gives nil if not a number
(defun nat-inv (n)
  (if (= n 0)
      0
    (let ((m (nat-inv (union1 n))))
      (if m
	  (if (= n (nat-as-set (1+ m)))
	      (1+ m)
	    nil)
	nil))))

(defun pair-inv (n)
  (let ((l nil)
	(r nil)
	(notapair nil))
    (dolist (a (num-to-list n))
      (let ((p (pair-inv a)))
	(if p
	    (cond ((equal (car p) 0) (push (cdr p) l))
		  ((equal (car p) 1) (push (cdr p) r))
		  (t (setq notapair t)))
	  (setq notapair t))))
    (if notapair
	nil
      (cons (list-to-num l) (list-to-num r)))))

(defun pair-p (n)
  (if (pair-inv n)
      t
    nil))

(defun tuple-inv (n m)
  (let ((k (nat-inv n)))
    (if k
	(let ((ar (make-array k :initial-element 0))
	      (notaktuple nil))
	  (dolist (a (num-to-list m))
	    (let ((p (pair-inv a)))
	      (if p
		  (let ((i (nat-inv (car p))))
		    (if (and i (< i k))
			(setf (aref ar i) (logior (cdr p) (aref ar i)))
		      (setq notaktuple t)))
		(setq notaktuple t))))
	  (if notaktuple
	      nil
	    ar))
      nil)))

(defun tuple-p (n m)
  (if (tuple-inv n m)
      t
    nil))

(defun ap (n m)
  (let ((r 0))
    (dolist (a (num-to-list n))
      (let ((p (pair-inv a)))
	(when (and p (equal (car p) m))
	  (setq r (setadjoin m (cdr p))))))
    r))

(defun proj0 (n)
  (ap n 0))

(defun proj1 (n)
  (ap n 1))

(defun inj1 (n)
  (setadjoin (list-to-num (mapcar #'inj1 (num-to-list n))) 0))

(defun inj0 (n)
  (list-to-num (mapcar #'inj1 (num-to-list n))))

(defun unj (n)
  (list-to-num (mapcar #'(lambda (z) (if (= (logand z 1) 1) (- z 1) z)) (num-to-list n))))

(defun setsum (n m)
  (list-to-num (append (mapcar #'inj0 (num-to-list n))
		       (mapcar #'inj1 (num-to-list m)))))

(defun setprod (n m)
  (let ((r nil)
	(nl (num-to-list n))
	(ml (num-to-list m)))
    (dolist (a nl)
      (dolist (b ml)
	(push (setsum a b) r)))
    (list-to-num r)))

(defun explist (doml codl)
  (if doml
      (let ((fl (explist (cdr doml) codl))
	    (rl nil))
	(dolist (a codl)
	  (dolist (f fl)
	    (push (binunion (list-to-num (mapcar #'(lambda (b) (setsum (car doml) b)) (num-to-list a))) f) rl)))
	rl)
    (list 0)))

(defun setexp (n m)
  (let* ((nl (num-to-list n))
	 (ml (num-to-list m))
	 (fl (explist nl ml)))
    (list-to-num fl)))

(defun ordsucc (n)
  (setadjoin n n))

(defun nat-as-set (k)
  (if (> k 0)
      (let ((n (nat-as-set (- k 1))))
	(ordsucc n))
    0))

; fail if not nats
(defun add-nat (n m)
  (let ((i (nat-inv n))
	(j (nat-inv m)))
    (if (and i j)
	(nat-as-set (+ i j))
      (throw 'fail (list 'cannotaddnats n m)))))

(defun mul-nat (n m)
  (let ((i (nat-inv n))
	(j (nat-inv m)))
    (if (and i j)
	(nat-as-set (* i j))
      (throw 'fail (list 'cannotmulnats n m)))))

(defun card0 (x) (equal x 0))
(defun card1 (x) (= (length (num-to-list x)) 1))
(defun card2 (x) (= (length (num-to-list x)) 2))
(defun card3 (x) (= (length (num-to-list x)) 3))
(defun card4 (x) (= (length (num-to-list x)) 4))
(defun card5 (x) (= (length (num-to-list x)) 5))
(defun card6 (x) (= (length (num-to-list x)) 6))

(defun sym (r) (let ((exc (cadr r)) (s t)) (dolist (x exc s) (unless (member (reverse x) exc :test #'equal) (setq s nil)))))

(defun reln-from (r x) (let ((def (car r)) (exc (cadr r))) (if def nil (subsetp (mapcar #'car exc) (num-to-list x) :test #'equal))))
(defun reln-to (r x) (let ((def (car r)) (exc (cadr r))) (if def nil (subsetp (mapcar #'cadr exc) (num-to-list x) :test #'equal))))
(defun reln-on (r x) (and (reln-from r x) (reln-to r x)))
(defun sym-reln-on (r x) (and (sym r) (reln-on r x)))

(defun compl (r) (list (not (car r)) (cadr r)))

(defun eval-pred (r x)
  (if (member x (cadr r) :test #'equal)
      (not (car r))
    (car r)))

(defun eval-reln (r x y)
  (if (member (list x y) (cadr r) :test #'equal)
      (not (car r))
    (car r)))

(defun clique (r u) (bd-forall u #'(lambda (x) (bd-forall u #'(lambda (y) (eval-reln r x y))))))

(defun all-clique (r v p) (bd-forall (power v) #'(lambda (u) (or (not (eval-pred p u)) (clique r u)))))

(defun head-spine (m &optional sp)
  (if (consp m)
      (head-spine (car m) (cons (cdr m) sp))
    (values m sp)))

(defun ev2 (m phi)
  (if (consp m)
      (cond ((eq (car m) 'NOT) (not (ev2 (cadr m) phi)))
	    ((eq (car m) 'IMP)
	     (let ((v1 (catch 'fail (ev2 (cadr m) phi))))
	       (if (eq v1 'fail)
		   (let ((v2 (ev2 (caddr m) phi)))
		     (if v2
			 t
		       (throw 'fail 'fail)))
		 (if v1
		     (ev2 (caddr m) phi)
		   t))))
	    ((eq (car m) 'OR)
	     (let ((v1 (catch 'fail (ev2 (cadr m) phi))))
	       (if (eq v1 'fail)
		   (let ((v2 (ev2 (caddr m) phi)))
		     (if v2
			 t
		       (throw 'fail 'fail)))
		 (if v1
		     t
		   (ev2 (caddr m) phi)))))
	    ((eq (car m) 'AND)
	     (let ((v1 (catch 'fail (ev2 (cadr m) phi))))
	       (if (eq v1 'fail)
		   (let ((v2 (ev2 (caddr m) phi)))
		     (if v2
			 (throw 'fail 'fail)
		       nil))
		 (if v1
		     (ev2 (caddr m) phi)
		   nil))))
;	    ((and (eq (car m) 'FORALL) (eq (caddr m) 'I))
;	     (let ((x (cadr m))
;		   (body (nth 3 m))
;		   (r t)
;		   (sl sets1))
;	       (loop while (and r sl) do
;		     (unless (ev2 body (acons x (car sl) phi))
;		       (setq r nil))
;		     (pop sl))
;	       r))
;	    ((and (eq (car m) 'EXISTS) (eq (caddr m) 'I))
;	     (let ((x (cadr m))
;		   (body (nth 3 m))
;		   (r nil)
;		   (sl sets1))
;	       (loop while (and (not r) sl) do
;		     (when (ev2 body (acons x (car sl) phi))
;		       (setq r t))
;		     (pop sl))
;	       r))
	    ((eq (car m) 'FORALL-IN)
	     (let ((x (cadr m))
		   (body (nth 3 m))
		   (r t)
		   (sl (num-to-list (ev3 (caddr m) phi))))
	       (loop while (and r sl) do
		     (unless (ev2 body (acons x (car sl) phi))
		       (setq r nil))
		     (pop sl))
	       r))
	    ((eq (car m) 'EXISTS-IN)
	     (let ((x (cadr m))
		   (body (nth 3 m))
		   (r nil)
		   (sl (num-to-list (ev3 (caddr m) phi))))
	       (loop while (and (not r) sl) do
		     (when (ev2 body (acons x (car sl) phi))
		       (setq r t))
		     (pop sl))
	       r))
	    ((eq (car m) 'FORALL-SUBQ)
	     (let* ((x (cadr m))
		    (body (nth 3 m))
		    (r t)
		    (bdl (num-to-list (ev3 (caddr m) phi)))
		    (sl (if (<= (length bdl) 16) (powerlist bdl) (throw 'fail 'fail))))
	       (loop while (and r sl) do
		     (unless (ev2 body (acons x (list-to-num (car sl)) phi))
		       (setq r nil))
		     (pop sl))
	       r))
	    ((eq (car m) 'EXISTS-SUBQ)
	     (let* ((x (cadr m))
		    (body (nth 3 m))
		    (r nil)
		    (bdl (num-to-list (ev3 (caddr m) phi)))
		    (sl (if (<= (length bdl) 16) (powerlist bdl) (throw 'fail 'fail))))
	       (loop while (and (not r) sl) do
		     (when (ev2 body (acons x (list-to-num (car sl)) phi))
		       (setq r t))
		     (pop sl))
	       r))
	    (t
	     (multiple-value-bind
	      (h sp)
	      (head-spine m)
	      (cond ((equal h "In") (mem (ev3 (car sp) phi) (ev3 (cadr sp) phi)))
		    ((or (equal h "=") (eq h 'EQ)) (= (ev3 (car sp) phi) (ev3 (cadr sp) phi)))
		    ((equal h "Subq") (subq (ev3 (car sp) phi) (ev3 (cadr sp) phi)))
		    ((equal h "disj") (disj (ev3 (car sp) phi) (ev3 (cadr sp) phi)))
		    ((equal h "Power_closed") (power-closed (ev3 (car sp) phi)))
		    ((equal h "TransSet") (TransSet (ev3 (car sp) phi)))
		    ((equal h "ordinal") (ordinal (ev3 (car sp) phi)))
		    ((equal h "nat_p") (nat-p (ev3 (car sp) phi)))
		    ((equal h "pair_p") (pair-p (ev3 (car sp) phi)))
		    ((equal h "clique")
		     (let ((a (assoc (car sp) phi)))
		       (if a
			   (ev2 '(FORALL-IN X1 X (FORALL-IN X2 X ((R . X1) . X2))) (acons 'R (cdr a) (acons 'X (ev3 (cadr sp) phi) nil)))
			 (throw 'fail 'fail))))
		    ((equal h "anticlique")
		     (let ((a (assoc (car sp) phi)))
		       (if a
			   (ev2 '(FORALL-IN X1 X (FORALL-IN X2 X (NOT ((R . X1) . X2)))) (acons 'R (cdr a) (acons 'X (ev3 (cadr sp) phi) nil)))
			 (throw 'fail 'fail))))
		    ((equal h "all_nonclique")
		     (let ((a (assoc (car sp) phi))
			   (b (assoc (cadr sp) phi)))
		       (if (and a b)
			   (ev2 '(FORALL-SUBQ X1 X (IMP (P . X1) (NOT (("clique" . R) . X1)))) (acons 'P (cdr a) (acons 'R (cdr b) (acons 'X (ev3 (caddr sp) phi) nil))))
			 (throw 'fail 'fail))))
		    ((equal h "ex_anticlique")
		     (let ((a (assoc (car sp) phi))
			   (b (assoc (cadr sp) phi)))
		       (if (and a b)
			   (ev2 '(EXISTS-SUBQ X1 X (AND (P . X1) (("anticlique" . R) . X1))) (acons 'P (cdr a) (acons 'R (cdr b) (acons 'X (ev3 (caddr sp) phi) nil))))
			 (throw 'fail 'fail))))
		    (t
		     (let ((a (assoc h phi)))
		       (if a
			   (let ((el (mapcar #'(lambda (x) (ev3 x phi)) sp)))
			     (member el (cdr a) :test #'equal))
			 (throw 'fail 'fail))))))))
    (throw 'fail 'fail)))

(defun ev3 (m phi)
  (let ((a (assoc m phi)))
    (if a
	(cdr a)
      (multiple-value-bind
       (h sp)
       (head-spine m)
       (cond ((equal h "Empty") 0)
	     ((equal h "Power") (power (ev3 (car sp) phi)))
	     ((equal h "Sing") (sing (ev3 (car sp) phi)))
	     ((equal h "UPair") (upair (ev3 (car sp) phi) (ev3 (cadr sp) phi)))
	     ((equal h "Union") (union1 (ev3 (car sp) phi)))
	     ((equal h "SetAdjoin") (setadjoin (ev3 (car sp) phi) (ev3 (cadr sp) phi)))
	     ((equal h "ordsucc") (let ((x (ev3 (car sp) phi))) (setadjoin x x)))
	     ((equal h "binunion") (binunion (ev3 (car sp) phi) (ev3 (cadr sp) phi)))
	     ((equal h "binintersect") (binintersect (ev3 (car sp) phi) (ev3 (cadr sp) phi)))
	     ((equal h "setminus") (setminus (ev3 (car sp) phi) (ev3 (cadr sp) phi)))
	     ((equal h "Unj") (unj (ev3 (car sp) phi)))
	     ((equal h "Inj0") (inj0 (ev3 (car sp) phi)))
	     ((equal h "Inj1") (inj1 (ev3 (car sp) phi)))
	     ((equal h "proj0") (proj0 (ev3 (car sp) phi)))
	     ((equal h "proj1") (proj1 (ev3 (car sp) phi)))
	     ((equal h "setsum") (setsum (ev3 (car sp) phi) (ev3 (cadr sp) phi)))
	     ((equal h "pair") (setsum (ev3 (car sp) phi) (ev3 (cadr sp) phi)))
	     ((equal h "ap") (ap (ev3 (car sp) phi) (ev3 (cadr sp) phi)))
	     ((equal h "setprod") (setprod (ev3 (car sp) phi) (ev3 (cadr sp) phi)))
	     ((equal h "setexp") (setexp (ev3 (car sp) phi) (ev3 (cadr sp) phi)))
	     (t (throw 'fail 'fail)))))))

(defun top ()
  (let ((n (member "-i" sb-ext:*posix-argv* :test #'equal))
	(o (member "-o" sb-ext:*posix-argv* :test #'equal)))
    (setq f (if n (open (cadr n) :direction :input) *standard-input*))
    (setq g (if o (open (cadr o) :direction :output :if-exists :supersede :if-does-not-exist :create) *standard-output*))
    (loop while (setq m (read f nil nil)) do
	  (setq phi (read f nil nil))
	  (let ((v (catch 'fail (ev2 m phi))))
	    (format g "~S~%" v)))
    (when n (close f))
    (when o (close g)))
  (exit))


  
    
